"""Filter edit."""

from __future__ import annotations

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QSizePolicy, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_button_icon import QtAndOrButton
from qtextra.widgets.qt_button_tag import QtTagButton
from qtextra.widgets.qt_layout_scroll import QtScrollableHLayoutWidget


class QtFilterEdit(QWidget):
    """Scrollable edit.."""

    evt_filters_changed = Signal(list, str)

    def __init__(
        self,
        parent: QWidget | None = None,
        placeholder: str = "Type in...",
        above: bool = False,
        flow: bool = False,
        enable_switch: bool = False,
    ):
        self.enable_switch = enable_switch
        super().__init__(parent)

        self.text_edit = hp.make_line_edit(
            self, placeholder=placeholder, func_changed=self.emit_current_filters, func=self.on_add
        )
        self.clear_action = hp.make_action(self, "clear", func=self.on_remove_all, tooltip="Remove all filters.")
        self.text_edit.addAction(self.clear_action, self.text_edit.ActionPosition.TrailingPosition)
        self.add_action = hp.make_action(
            self, "add", func=self.on_add_action, tooltip="Add currently entered text as a filter."
        )
        self.text_edit.addAction(self.add_action, self.text_edit.ActionPosition.TrailingPosition)

        self.switch_toggle = QtAndOrButton(auto_connect=True, state=True)
        self.switch_toggle.set_normal()
        self.switch_toggle.evt_toggled.connect(self.emit_current_filters)
        if not enable_switch:
            self.switch_toggle.hide()

        if flow:
            self._filter_layout = hp.make_flow_layout(margin=0, horizontal_spacing=1, vertical_spacing=1)
        else:
            self._filter_layout = QtScrollableHLayoutWidget(self)
            self._filter_layout.set_min_height(self.text_edit.height())
        self._n = self._filter_layout.count()

        self._main_layout = hp.make_form_layout(margin=0)
        if above:
            self._main_layout.addRow(self._filter_layout)
            self._main_layout.addRow(
                hp.make_h_layout(self.text_edit, self.switch_toggle, spacing=1, margin=0, stretch_id=(0,))
            )
        else:
            self._main_layout.addRow(
                hp.make_h_layout(self.text_edit, self.switch_toggle, spacing=1, margin=0, stretch_id=(0,))
            )
            self._main_layout.addRow(self._filter_layout)
        self.setLayout(self._main_layout)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

    def on_add_action(self) -> None:
        """This is a workaround some minor issues with action being triggered twice."""
        with hp.qt_signals_blocked(self.text_edit):
            self.on_add()

    def on_add(self) -> None:
        """Add filter."""
        text = self.text_edit.text()
        self.add_filter(text)

    def add_filter(self, text: str) -> None:
        """Add filter."""
        filters = self.get_filters()
        if not text or text in filters:
            return
        button = QtTagButton(text, text, allow_selected=False, action_type="delete", action_icon="cross")
        button.evt_action.connect(self.on_remove)
        button.evt_clicked.connect(self.on_update)
        self._filter_layout.insertWidget(0, button)
        self.evt_filters_changed.emit(self.get_filters(), self.mode)
        self.text_edit.setText("")

    def on_update(self, text: str) -> None:
        """Update filter."""
        self.text_edit.setText(text)
        self.text_edit.setFocus()
        self.on_remove(text)

    def on_remove(self, text: str) -> None:
        """Remove filter."""
        filters = self.get_filters()
        index = filters.index(text)
        self._filter_layout.removeWidgetOrLayout(index)
        self.evt_filters_changed.emit(self.get_filters(), self.mode)

    def on_remove_all(self) -> None:
        """Remove all filters."""
        while self._filter_layout.count() != self._n:
            self._filter_layout.removeWidgetOrLayout(0)
        self.evt_filters_changed.emit([], self.mode)

    def get_filters(self, current: bool = False) -> list[str]:
        """Get list of currently selected filters."""
        tags = []
        for i in range(self._filter_layout.count() - self._n):
            widget = self._filter_layout.get_widget(i)
            if widget:
                tags.append(widget.text)
        if current:
            text = self.text_edit.text()
            if text and text not in tags:
                tags.append(text)
        return tags

    @property
    def mode(self) -> str | None:
        """Get mode."""
        if not self.enable_switch:
            return None
        return "AND" if self.switch_toggle.state else "OR"

    def emit_current_filters(self) -> None:
        """Emit current filters."""
        self.evt_filters_changed.emit(self.get_filters(current=True), self.mode)

    # Alias methods to offer Qt-like interface
    getFilters = get_filters
    addFilter = add_filter
    onAdd = on_add
    onRemove = on_remove
    onRemoveAll = on_remove_all
    emitCurrentFilters = emit_current_filters


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(horz=False)
    frame.setMinimumSize(600, 600)
    ha.addWidget(QtFilterEdit())
    ha.addWidget(QtFilterEdit(above=True))
    ha.addWidget(QtFilterEdit(flow=True))
    ha.addWidget(QtFilterEdit(enable_switch=True))

    frame.show()
    sys.exit(app.exec_())



"""Custom image button class."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from copy import deepcopy
from functools import partial

import qtawesome
from qtpy.QtCore import (  # type: ignore[attr-defined]
    QEasingCurve,
    QEvent,
    QPoint,
    QPointF,
    QPropertyAnimation,
    QRectF,
    Qt,
    Signal,
    Slot,
)
from qtpy.QtGui import QBrush, QColor, QFont, QPainter
from qtpy.QtWidgets import QGraphicsOpacityEffect, QHBoxLayout, QLabel, QPushButton, QToolTip, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.assets import get_icon
from qtextra.config import THEMES
from qtextra.widgets._qta_mixin import QtaMixin

INDICATOR_TYPES = {"success": "success", "warning": "warning", "active": "progress"}


class QtImagePushButton(QPushButton, QtaMixin):
    """Image button."""

    evt_click = Signal(QPushButton)
    evt_right_click = Signal(QPushButton)
    has_right_click: bool = False
    count: int = 0
    count_enabled: bool = False

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        self._icon_color = kwargs.pop("icon_color_override", None)
        super().__init__()
        self.setProperty("transparent", False)
        self.transparent = False
        with suppress(RuntimeError):
            THEMES.evt_theme_icon_changed.connect(self._update_qta)

    def set_count(self, count: int, enabled: bool = True) -> None:
        """Enable count indicator."""
        self.count = count
        self.count_enabled = enabled

    def setText(self, text: str) -> None:  # type: ignore[override]
        """Override text."""
        self.setProperty("with_text", True)
        super().setText(text)

    def set_transparent(self, transparent: bool) -> None:
        """Set transparency."""
        from qtextra.helpers import polish_widget

        self.transparent = transparent
        polish_widget(self)

    def mousePressEvent(self, evt: QEvent) -> None:  # type: ignore[override]
        """Mouse press event."""
        if evt.button() == Qt.MouseButton.RightButton:  # type: ignore[attr-defined]
            self.on_right_click()
        elif evt.button() == Qt.MouseButton.LeftButton:  # type: ignore[attr-defined]
            self.on_click()
        super().mousePressEvent(evt)  # type: ignore[arg-type]

    def set_toggle_qta(self, name: str, checked_name: str, connect: bool = True, **kwargs: ty.Any) -> None:
        """Set changeable icon."""
        checked_kwargs = deepcopy(kwargs)

        name, kwargs_ = get_icon(name)
        kwargs.update(kwargs_)
        checked_name, checked_kwargs_ = get_icon(checked_name)
        checked_kwargs.update(checked_kwargs_)
        self._qta_data = (name, kwargs)
        self._checked_qta_data = (checked_name, checked_kwargs)
        color_ = kwargs.pop("color", None)
        color = color_ or self._icon_color or THEMES.get_hex_color("icon")
        icon = qtawesome.icon(
            checked_name if self.isChecked() else name,
            **self._checked_qta_data[1] if self.isChecked() else self._qta_data[1],
            color=color,
        )
        self.setIcon(icon)
        if connect:
            self.toggled.connect(self._on_toggle)

    def _on_toggle(self) -> None:
        """Update icons."""
        assert self._qta_data and self._checked_qta_data, "No qta data set."
        name = self._checked_qta_data[0] if self.isChecked() else self._qta_data[0]
        self._set_qta_icon(name, **self._checked_qta_data[1] if self.isChecked() else self._qta_data[1])

    def on_click(self) -> None:
        """Click event."""
        self.evt_click.emit(self)

    def on_right_click(self) -> None:
        """Right click event."""
        self.evt_right_click.emit(self)

    def connect_to_right_click(self, func: ty.Callable) -> None:
        """Connect function right right-click.

        It is not possible to check whether a function is connected to a signal so its better to use this function to
        connect via this function which leaves behind a flag so the paint event will add rectangle to the edge so the
        user knows there is a right-click menu available.
        """
        self.evt_right_click.connect(func)
        self.has_right_click = True
        hp.set_properties(self, {"right_click": True})

    def paintEvent(self, *args: ty.Any) -> None:
        """Paint event."""
        super().paintEvent(*args)
        paint = QPainter(self)
        if self.has_right_click:
            width = self.rect().width() / 6
            radius = self.rect().width() / 8
            x = self.rect().width() - width
            y = self.rect().height() - width
            color = THEMES.get_hex_color("success")
            paint.setPen(QColor(color))
            paint.setBrush(QColor(color))
            paint.drawEllipse(QPointF(x, y), radius, radius)
        if self.count_enabled:
            # add text
            text = "9+" if self.count > 9 else str(self.count)

            color = THEMES.get_hex_color("icon")
            paint.setPen(QColor(color))
            paint.setBrush(QColor(color))
            radius = self.rect().width() / 6
            x = self.rect().x()
            y = self.rect().y()
            rect = QRectF(x, y, radius * 4, radius * 4)
            font: QFont = paint.font()
            font.setPointSize(12)
            font.setBold(True)
            paint.setFont(font)
            paint.drawText(rect, Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft, text)

    # Alias methods to offer Qt-like interface
    _onToggle = _on_toggle
    setCount = set_count
    setTransparent = set_transparent
    setToggleQta = set_toggle_qta
    onClick = on_click
    onRightClick = on_right_click
    connectToRightClick = connect_to_right_click


class QtTogglePushButton(QtImagePushButton):
    """Toggle button."""

    evt_toggled = Signal(bool)

    ICON_ON: str = ""
    ICON_OFF: str = ""

    _state: bool = False

    def __init__(self, *args, state: bool = False, auto_connect: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        if auto_connect:
            self.auto_connect()
        self.state = state

    @property
    def state(self) -> bool:
        """Get state."""
        return self._state

    @state.setter
    def state(self, value: bool) -> None:
        changed = value != self._state
        self._state = value
        self.set_qta(self.ICON_ON if value else self.ICON_OFF)
        if changed:
            self.evt_toggled.emit(value)

    def set_state(self, state: bool, trigger: bool = True) -> None:
        """Set state."""
        with hp.qt_signals_blocked(self, block_signals=not trigger):
            self.state = state

    def auto_connect(self) -> None:
        """Automatically connect."""
        self.evt_click.connect(self.toggle_state)

    def toggle_state(self) -> None:
        """Toggle state."""
        self.state = not self.state

    # Alias methods to offer Qt-like interface
    setState = set_state
    autoConnect = auto_connect
    toggleState = toggle_state


class QtAnimationPlayButton(QtTogglePushButton):
    """Play button with multiple states to indicate current state."""

    ICON_ON = "stop"
    ICON_OFF = "start"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def playing(self) -> bool:
        """Get playing state."""
        return self.state

    @playing.setter
    def playing(self, state: bool) -> None:
        self.state = state


class QtPauseButton(QtTogglePushButton):
    """Play button with multiple states to indicate current state."""

    ICON_ON = "start"
    ICON_OFF = "pause"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def paused(self) -> bool:
        """Get playing state."""
        return self.state

    @paused.setter
    def paused(self, state: bool) -> None:
        self.state = state


class QtLockButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "lock_closed"
    ICON_OFF = "lock_open"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def locked(self) -> bool:
        """Get playing state."""
        return self.state

    @locked.setter
    def locked(self, state: bool) -> None:
        self.state = state


class QtThemeButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "dark_theme"
    ICON_OFF = "light_theme"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def dark(self) -> bool:
        """Get playing state."""
        return self.state

    @dark.setter
    def dark(self, state: bool) -> None:
        self.state = state


class QtAndOrButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "and"
    ICON_OFF = "or"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def dark(self) -> bool:
        """Get playing state."""
        return self.state

    @dark.setter
    def dark(self, state: bool) -> None:
        self.state = state


class QtExpandButton(QtTogglePushButton):
    """Button that has chevron point up or down."""

    ICON_ON = "chevron_up"
    ICON_OFF = "chevron_down"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def expanded(self) -> bool:
        """Get state."""
        return self.state

    @expanded.setter
    def expanded(self, state: bool) -> None:
        self.state = state


class QtSortButton(QtTogglePushButton):
    """Button that has chevron point up or down."""

    ICON_ON = "sort_ascending"
    ICON_OFF = "sort_descending"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)


class QtToggleButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "toggle_on"
    ICON_OFF = "toggle_off"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property  # type: ignore[override]
    def toggled(self) -> bool:
        """Get toggle state."""
        return self.state

    @toggled.setter
    def toggled(self, state: bool) -> None:
        self.state = state


class QtVerticalDirectionButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "long_arrow_up"
    ICON_OFF = "long_arrow_down"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def up(self) -> bool:
        """Get toggle state."""
        return self.state

    @up.setter
    def up(self, state: bool) -> None:
        self.state = state

    @property
    def down(self) -> bool:
        """Get toggle state."""
        return not self.up


class QtHorizontalDirectionButton(QtTogglePushButton):
    """Lock button with open/closed state to indicate current state."""

    ICON_ON = "long_arrow_right"
    ICON_OFF = "long_arrow_left"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def right(self) -> bool:
        """Get toggle state."""
        return self.state

    @right.setter
    def right(self, state: bool) -> None:
        self.state = state

    @property
    def left(self) -> bool:
        """Get toggle state."""
        return not self.left


class QtVisibleButton(QtTogglePushButton):
    """Lock button with shown/hidden icon."""

    ICON_ON = "visible_on"
    ICON_OFF = "visible_off"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def visible(self) -> bool:
        """Get toggle state."""
        return self.state

    @visible.setter
    def visible(self, state: bool) -> None:
        self.state = state

    @property
    def hidden(self) -> bool:
        """Get toggle state."""
        return not self.visible


class QtPinButton(QtTogglePushButton):
    """Lock button with shown/hidden icon."""

    ICON_ON = "pin_on"
    ICON_OFF = "pin_off"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def pin(self) -> bool:
        """Get toggle state."""
        return self.state

    @pin.setter
    def pin(self, state: bool) -> None:
        self.state = state


class QtFullscreenButton(QtTogglePushButton):
    """Lock button with shown/hidden icon."""

    ICON_ON = "fullscreen"
    ICON_OFF = "maximize"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def fullscreen(self) -> bool:
        """Get toggle state."""
        return self.state

    @fullscreen.setter
    def fullscreen(self, state: bool) -> None:
        self.state = state


class QtMinimizeButton(QtTogglePushButton):
    ICON_ON = "minimize"
    ICON_OFF = "maximize"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

    @property
    def minimized(self) -> bool:
        """Get toggle state."""
        return self.state

    @minimized.setter
    def minimized(self, state: bool) -> None:
        self.state = state


class QtBoolButton(QtTogglePushButton):
    """Boolean button."""

    ICON_ON = "true"
    ICON_OFF = "false"

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)


class QtMultiStatePushButton(QtImagePushButton):
    """Base class for a multi-state button where options are shown as a QMenu."""

    DEFAULT_STATE: str = ""
    STATE_TO_ICON: ty.ClassVar[dict[str, str]]
    STATE_TO_OPTION: ty.ClassVar[dict[str, str]]

    evt_changed = Signal(str)

    _state: str = ""
    _menu: ty.Optional[QWidget]

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.setMouseTracking(True)
        if self.DEFAULT_STATE:
            self.state = self.DEFAULT_STATE

    @property
    def state(self) -> str:
        """Get playing state."""
        return self._state

    @state.setter
    def state(self, state: str) -> None:
        self._state = state
        self.set_qta(self.STATE_TO_ICON[state])
        self.evt_changed.emit(state)

    def set_state(self, state: bool) -> None:
        """Set state."""
        self.state = state

    def set_and_show_menu(self) -> None:
        """Set menu."""
        menu = hp.make_menu(self)
        for state, label in self.STATE_TO_OPTION.items():
            hp.make_menu_item(
                self,
                label,
                icon=self.STATE_TO_ICON[state],
                func=partial(self.set_state, state=state),
                menu=menu,
            )
        self._menu = menu
        hp.show_below_widget(menu, self, x_offset=20)

    def enterEvent(self, event: QEvent) -> None:  # type: ignore[override]
        """Event."""
        self.set_and_show_menu()
        super().enterEvent(event)  # type: ignore[arg-type]

    def leaveEvent(self, event: QEvent) -> None:  # type: ignore[override]
        """Event."""
        self._menu.close()  # type: ignore[union-attr]
        self._menu = None
        super().leaveEvent(event)

    # Alias methods to offer Qt-like interface
    setAndShowMenu = set_and_show_menu
    setState = set_state


class QtPriorityButton(QtMultiStatePushButton):
    """Priority button."""

    DEFAULT_STATE = "normal"
    STATE_TO_ICON: ty.ClassVar[dict] = {"low": "priority_low", "normal": "priority_normal", "high": "priority_high"}
    STATE_TO_OPTION: ty.ClassVar[dict] = {"low": "Low", "normal": "Normal", "high": "High"}

    @property
    def priority(self) -> str:
        """Get playing state."""
        return self.state

    @priority.setter
    def priority(self, state: str) -> None:
        self.state = state


class QtStateButton(QtMultiStatePushButton):
    """State button."""

    DEFAULT_STATE = "info"
    STATE_TO_ICON: ty.ClassVar[dict] = {
        "success": "success",
        "debug": "debug",
        "info": "info",
        "warning": "warning",
        "error": "error",
    }
    STATE_TO_OPTION: ty.ClassVar[dict] = {
        "success": "Success",
        "debug": "Debug",
        "info": "Info",
        "warning": "Warning",
        "error": "Error",
    }


class QtMultiThemeButton(QtMultiStatePushButton):
    """Theme button button."""

    ICON_ON = "dark_theme"
    ICON_OFF = "light_theme"

    DEFAULT_STATE = "light"
    STATE_TO_ICON: ty.ClassVar[dict] = {
        "light": "light_theme",
        "dark": "dark_theme",
    }
    STATE_TO_OPTION: ty.ClassVar[dict] = {
        "light": "Light",
        "dark": "Dark",
    }

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        THEMES.evt_theme_added.connect(self._update_states)

    def _update_states(self) -> None:
        """Update states."""
        for theme in THEMES.available_themes():
            if theme not in self.STATE_TO_OPTION:
                self.STATE_TO_OPTION[theme] = theme.capitalize()
                self.STATE_TO_ICON[theme] = theme


class QtToolbarPushButton(QtImagePushButton):
    """Image button."""

    START_OPACITY = 1.0
    END_OPACITY = 0.2
    PULSE_RATE = 1000
    N_LOOPS = 5

    indicator: str = ""
    _text: str = ""

    panel_widget: QWidget | None = None
    about_widget: QWidget | None = None

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.setMouseTracking(True)

        self.opacity = QGraphicsOpacityEffect()
        self.opacity.setOpacity(1.0)
        self.setGraphicsEffect(self.opacity)
        self.opacity_anim = QPropertyAnimation(self.opacity, b"opacity", self)
        self.opacity_anim.currentLoopChanged.connect(self._loop_update)
        self.opacity_anim.finished.connect(self.stop_pulse)

        self.evt_click.connect(self.stop_pulse)

    def setToolTip(self, text: str) -> None:  # type: ignore[override]
        """Override tooltip."""
        self._text = text

    def _get_position(self) -> QPoint:
        rect = self.rect()
        pos = self.mapToGlobal(rect.topRight())
        pos -= QPoint(0, 22)
        return pos

    def event(self, evt: QEvent) -> bool:  # type: ignore[override]
        """Override event handler to quickly display/hide tooltip."""
        if evt.type() == QEvent.Type.Enter:
            QToolTip.showText(self._get_position(), self._text)
            evt.ignore()
        elif evt.type() == QEvent.Type.Leave:
            QToolTip.hideText()
        return super().event(evt)

    @Slot(int)  # type: ignore[misc]
    def _loop_update(self, loop: int) -> None:
        """Reverse pulse direction for nicer visual effect."""
        start, end = (self.START_OPACITY, self.END_OPACITY) if loop % 2 == 0 else (self.END_OPACITY, self.START_OPACITY)
        self.opacity_anim.setStartValue(start)
        self.opacity_anim.setEndValue(end)

    @Slot(str)  # type: ignore[misc]
    @Slot(str, str)  # type: ignore[misc]
    def set_indicator(self, indicator_type: str, about: ty.Optional[str] = None) -> None:
        """Set indicator type."""
        assert indicator_type in [
            "",
            "success",
            "warning",
            "active",
        ], f"Cannot use `{indicator_type}` type of indicator."
        if not self.isCheckable():
            indicator_type = ""
        self.indicator = indicator_type
        self.start_pulse() if self.indicator else self.stop_pulse()

    @property
    def indicator_color(self) -> QColor:
        """Indicator color."""
        return QColor(THEMES.get_hex_color(INDICATOR_TYPES[self.indicator]))

    @property
    def edge_color(self) -> QColor:
        """Edge color."""
        return QColor(THEMES.get_hex_color(INDICATOR_TYPES[self.indicator]))

    def paintEvent(self, *args: ty.Any) -> None:
        """Paint event."""
        # default paint
        QPushButton.paintEvent(self, *args)

        if self.indicator and not self.isChecked():
            width = int(self.rect().width() / 6)
            radius = int(self.rect().width() / 10)
            pos = QPoint(self.rect().width() - width, width)

            paint = QPainter(self)
            pen = paint.pen()
            pen.setColor(self.edge_color)
            brush = QBrush(self.indicator_color)
            paint.setBrush(brush)
            paint.setPen(pen)
            paint.drawEllipse(pos, radius, radius)

    def start_pulse(self) -> None:
        """Start pulsating."""
        if self.indicator and not self.isChecked():
            self.opacity_anim.setEasingCurve(QEasingCurve.Type.Linear)
            self.opacity_anim.setDuration(self.PULSE_RATE)
            self.opacity_anim.setStartValue(self.START_OPACITY)
            self.opacity_anim.setEndValue(self.END_OPACITY)
            self.opacity_anim.setLoopCount(self.N_LOOPS)
            self.opacity_anim.start()

    def stop_pulse(self) -> None:
        """Stop pulsating."""
        self.opacity_anim.stop()
        self.opacity.setOpacity(1.0)

    # Alias methods to offer Qt-like interface
    setIndicator = set_indicator
    _getPosition = _get_position
    startPulse = start_pulse
    stopPulse = stop_pulse


class QtLabelledToolbarPushButton(QWidget):
    """Push button with label."""

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)

        self.image_btn = QtToolbarPushButton()
        self.label = QLabel()
        layout = QVBoxLayout(self)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.setSpacing(2)
        layout.addWidget(self.image_btn, alignment=Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(self.label, alignment=Qt.AlignmentFlag.AlignCenter)

    def set_label(self, text: str) -> None:
        """Set label."""
        self.label.setText(text)

    def set_qta(self, name: str, **kwargs: ty.Any) -> None:
        """Set icon."""
        self.image_btn.set_qta(name, **kwargs)

    # Alias methods to offer Qt-like interface
    setQta = set_qta
    setLabel = set_label


if __name__ == "__main__":  # pragma: no cover
    import sys

    val = True

    def _main() -> None:  # type: ignore[no-untyped-def]
        from qtextra.assets import QTA_MAPPING
        from qtextra.utils.dev import qdev, qmain

        app, frame, va = qmain(False)
        frame.setMinimumSize(600, 600)

        dev = qdev(frame)
        va.addWidget(dev)

        ha = QHBoxLayout()
        va.addLayout(ha)

        lay = QVBoxLayout()
        ha.addLayout(lay)

        btn2 = QtImagePushButton(parent=frame)
        btn2.setObjectName("info")
        lay.addWidget(btn2)

        lay.addWidget(QtAnimationPlayButton(parent=frame, auto_connect=True))
        lay.addWidget(QtPauseButton(parent=frame, auto_connect=True))
        lay.addWidget(QtLockButton(parent=frame, auto_connect=True))
        lay.addWidget(QtVerticalDirectionButton(parent=frame, auto_connect=True))
        lay.addWidget(QtHorizontalDirectionButton(parent=frame, auto_connect=True))
        lay.addWidget(QtVisibleButton(parent=frame, auto_connect=True))
        lay.addWidget(QtToggleButton(parent=frame, auto_connect=True))
        lay.addWidget(QtExpandButton(parent=frame, auto_connect=True))
        lay.addWidget(QtSortButton(parent=frame, auto_connect=True))
        lay.addWidget(QtPinButton(parent=frame, auto_connect=True))
        lay.addWidget(QtFullscreenButton(parent=frame, auto_connect=True))
        lay.addWidget(QtMinimizeButton(parent=frame, auto_connect=True))
        lay.addWidget(QtThemeButton(parent=frame, auto_connect=True))
        lay.addWidget(QtBoolButton(parent=frame, auto_connect=True))
        lay.addWidget(QtAndOrButton(parent=frame, auto_connect=True))

        # multi-state
        lay.addWidget(QtPriorityButton(parent=frame, auto_connect=True))
        lay.addWidget(QtStateButton(parent=frame, auto_connect=True))
        lay.addWidget(QtMultiThemeButton(parent=frame, auto_connect=True))

        ha.addWidget(hp.make_v_line())

        lay = QVBoxLayout()
        for i, (name, qta_name) in enumerate(QTA_MAPPING.items()):
            btn = QtImagePushButton()
            if i % 2 == 0:
                btn.connect_to_right_click(lambda: print("Right click"))
            if i % 3 == 0:
                btn.set_count(i)
            btn.set_qta(qta_name, scale_factor=1)
            btn.setToolTip(f"{name} :: {qta_name}")
            lay.addWidget(btn)
            if i % 10 == 0:
                ha.addLayout(lay)
                lay = QVBoxLayout()
        ha.addWidget(hp.make_v_line())

        lay = QVBoxLayout()
        for i, (name, qta_name) in enumerate(QTA_MAPPING.items()):
            btn = QtToolbarPushButton()
            btn.set_qta(qta_name)
            btn.set_large()
            btn.setToolTip(f"{name} :: {qta_name}")
            lay.addWidget(btn)
            if i % 10 == 0:
                ha.addLayout(lay)
                lay = QVBoxLayout()
        ha.addWidget(hp.make_v_line())

        lay = QVBoxLayout()
        for i, (name, qta_name) in enumerate(QTA_MAPPING.items()):
            btn = QtLabelledToolbarPushButton()
            btn.set_qta(qta_name)
            btn.set_label(name)
            btn.setToolTip(f"{name} :: {qta_name}")
            lay.addWidget(btn)
            if i == 10:
                ha.addLayout(lay)
                break

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Horizontal and Vertical lines."""

import typing as ty

from qtpy.QtWidgets import QFrame, QWidget

import qtextra.helpers as hp

# TODO: the vertical line with text is currently not looking great since the label is not properly rotated


class QtHorzLine(QFrame):
    """Horizontal line."""

    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setFrameShadow(QFrame.Shadow.Plain)


class QtHorzLineWithText(QWidget):
    """Horizontal line with text."""

    def __init__(self, parent: QWidget, label: str, bold: bool = False, position: str = "center", **kwargs: ty.Any):
        super().__init__(parent=parent)

        self.label = hp.make_label(parent, label, bold=bold, **kwargs)
        if position == "center":
            widgets = (QtHorzLine(parent), self.label, QtHorzLine(parent))
            stretch_ids = (0, 2)
        elif position == "left":
            widgets = (self.label, QtHorzLine(parent))
            stretch_ids = (1,)
        else:
            widgets = (QtHorzLine(parent), self.label)
            stretch_ids = (0,)
        hp.make_h_layout(*widgets, stretch_id=stretch_ids, spacing=2, margin=1, parent=self)

    def setText(self, text: str) -> None:
        """Set text of the label."""
        self.label.setText(text)


class QtVertLine(QFrame):
    """Vertical line."""

    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)
        self.setFrameShape(QFrame.Shape.VLine)
        self.setFrameShadow(QFrame.Shadow.Plain)


class QtVertLineWithText(QWidget):
    """Vertical line with text."""

    def __init__(self, parent: QWidget, label: str, bold: bool = False, position: str = "center", **kwargs: ty.Any):
        super().__init__(parent=parent)

        self.label = hp.make_label(parent, label, bold=bold, vertical=True, **kwargs)
        if position == "center":
            widgets = (QtVertLine(parent), self.label, QtVertLine(parent))
            stretch_ids = (0, 2)
        elif position == "left":
            widgets = (self.label, QtVertLine(parent))
            stretch_ids = (1,)
        else:
            widgets = (QtVertLine(parent), self.label)
            stretch_ids = (0,)
        hp.make_v_layout(
            *widgets,
            stretch_id=stretch_ids,
            spacing=2,
            margin=1,
            parent=self,
        )


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False, dev=True)
    frame.setMinimumSize(600, 600)

    ha.addWidget(QtHorzLineWithText(frame, "Test 1"))
    ha.addWidget(QtVertLineWithText(frame, "Test 1"))

    frame.show()
    sys.exit(app.exec_())



"""Qt flow layout."""

from __future__ import annotations

from qtpy.QtCore import (
    QEasingCurve,
    QEvent,
    QObject,
    QParallelAnimationGroup,
    QPoint,
    QPropertyAnimation,
    QRect,
    QSize,
    Qt,
    QTimer,
)
from qtpy.QtWidgets import QLayout, QWidget, QWidgetItem
from superqt import QFlowLayout


class QtAnimatedFlowLayout(QLayout):
    """Flow layout."""

    def __init__(self, parent=None, use_animation=False, tight=False):
        """
        Parameters
        ----------
        parent:
            parent window or layout
        use_animation: bool
            whether to add moving animation
        tight: bool
            whether to use the tight layout when widgets are hidden
        """
        super().__init__(parent)
        self._items = []  # type: List[QLayoutItem]
        self._animations = []  # type: List[QPropertyAnimation]
        self._animation_group = QParallelAnimationGroup(self)
        self._y_spacing = 10
        self._x_spacing = 10
        self.duration = 300
        self.easing_curve = QEasingCurve.Type.Linear
        self.use_animation = use_animation
        self.tight = tight
        self._debounce_timer = QTimer(self)
        self._debounce_timer.setSingleShot(True)
        self._debounce_timer.timeout.connect(lambda: self._doLayout(self.geometry(), True))
        self._parent = None
        self._has_event_filter = False

    def addItem(self, item):
        self._items.append(item)

    def insertItem(self, index, item):
        self._items.insert(index, item)

    def addWidget(self, w):
        super().addWidget(w)
        self._onWidgetAdded(w)

    def insertWidget(self, index, w):
        self.insertItem(index, QWidgetItem(w))
        self.addChildWidget(w)
        self._onWidgetAdded(w, index)

    def _onWidgetAdded(self, w, index=-1):
        if not self._has_event_filter:
            if w.parent():
                self._parent = w.parent()
                w.parent().installEventFilter(self)
            else:
                w.installEventFilter(self)

        if not self.use_animation:
            return

        ani = QPropertyAnimation(w, b"geometry")
        ani.setEndValue(QRect(QPoint(0, 0), w.size()))
        ani.setDuration(self.duration)
        ani.setEasingCurve(self.easing_curve)
        w.setProperty("flowAni", ani)
        self._animation_group.addAnimation(ani)

        if index == -1:
            self._animations.append(ani)
        else:
            self._animations.insert(index, ani)

    def setAnimation(self, duration: int, ease=QEasingCurve.Type.Linear):
        """Set the moving animation.

        Parameters
        ----------
        duration: int
            the duration of animation in milliseconds

        ease: QEasingCurve
            the easing curve of animation
        """
        if not self.use_animation:
            return
        self.duration = duration
        self.easing_curve = ease
        for animation in self._animations:
            animation.setDuration(duration)
            animation.setEasingCurve(ease)

    def count(self):
        return len(self._items)

    def itemAt(self, index: int):
        if 0 <= index < len(self._items):
            return self._items[index]

        return None

    def takeAt(self, index: int):
        if 0 <= index < len(self._items):
            item = self._items[index]  # type: QLayoutItem
            ani = item.widget().property("flowAni")
            if ani:
                self._animations.remove(ani)
                self._animation_group.removeAnimation(ani)
                ani.deleteLater()

            return self._items.pop(index).widget()

        return None

    def removeWidget(self, widget):
        """Remove widget from layout."""
        for i, item in enumerate(self._items):
            if item.widget() is widget:
                return self.takeAt(i)

    def removeAllWidgets(self):
        """Remove all widgets from layout."""
        while self._items:
            self.takeAt(0)

    def takeAllWidgets(self):
        """Remove all widgets from layout and delete them."""
        while self._items:
            w = self.takeAt(0)
            if w:
                w.deleteLater()

    def expandingDirections(self):
        """Get the expanding direction."""
        return Qt.Orientation(0)

    def hasHeightForWidth(self):
        """Check if the layout has height for width."""
        return True

    def heightForWidth(self, width: int):
        """Get the minimal height according to width."""
        return self._doLayout(QRect(0, 0, width, 0), False)

    def setGeometry(self, rect: QRect):
        super().setGeometry(rect)
        if self.use_animation:
            self._debounce_timer.start(80)
        else:
            self._doLayout(rect, True)

    def sizeHint(self):
        """Get the size hint of the layout."""
        return self.minimumSize()

    def minimumSize(self):
        """Get the minimal size of the layout."""
        size = QSize()
        for item in self._items:
            size = size.expandedTo(item.minimumSize())
        m = self.contentsMargins()
        size += QSize(m.left() + m.right(), m.top() + m.bottom())
        return size

    def setVerticalSpacing(self, spacing: int):
        """Set vertical spacing between widgets."""
        self._y_spacing = spacing

    def verticalSpacing(self):
        """Get vertical spacing between widgets."""
        return self._y_spacing

    def setHorizontalSpacing(self, spacing: int):
        """Set horizontal spacing between widgets."""
        self._x_spacing = spacing

    def horizontalSpacing(self):
        """Get horizontal spacing between widgets."""
        return self._x_spacing

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if obj in [w.widget() for w in self._items] and event.type() == QEvent.Type.ParentChange:
            self._parent = obj.parent()
            obj.parent().installEventFilter(self)
            self._has_event_filter = True

        if obj == self._parent and event.type() == QEvent.Type.Show:
            self._doLayout(self.geometry(), True)
            self._has_event_filter = True

        return super().eventFilter(obj, event)

    def _doLayout(self, rect: QRect, move: bool):
        """Adjust widgets position according to the window size."""
        animation_restart = False
        margin = self.contentsMargins()
        x = rect.x() + margin.left()
        y = rect.y() + margin.top()
        row_height = 0
        space_x = self.horizontalSpacing()
        space_y = self.verticalSpacing()

        for i, item in enumerate(self._items):
            if item.widget() and not item.widget().isVisible() and self.tight:
                continue

            nextX = x + item.sizeHint().width() + space_x

            if nextX - space_x > rect.right() - margin.right() and row_height > 0:
                x = rect.x() + margin.left()
                y = y + row_height + space_y
                nextX = x + item.sizeHint().width() + space_x
                row_height = 0

            if move:
                target = QRect(QPoint(x, y), item.sizeHint())
                if not self.use_animation:
                    item.setGeometry(target)
                elif target != self._animations[i].endValue():
                    self._animations[i].stop()
                    self._animations[i].setEndValue(target)
                    animation_restart = True

            x = nextX
            row_height = max(row_height, item.sizeHint().height())

        if self.use_animation and animation_restart:
            self._animation_group.stop()
            self._animation_group.start()
        return y + row_height + margin.bottom() - rect.y()


class QtFlowLayout(QFlowLayout):
    def __init__(self, parent: QWidget | None = None, margin: int = 0, spacing: int = -1):
        super().__init__(parent)
        if isinstance(margin, int):
            margin = (margin, margin, margin, margin)
        if margin:
            self.setContentsMargins(*margin)
        if spacing >= 0:
            self.setSpacing(spacing)

    def get_widget(self, index: int) -> QWidget | None:
        """Get widget at position."""
        item = self.itemAt(index)
        if not item:
            return None
        widget = item.widget()
        return widget

    def removeWidgetOrLayout(self, index: int) -> None:
        """Remove widget or layout based on index position."""
        widget = self.get_widget(index)
        if widget:
            self.removeWidget(widget)
            widget.deleteLater()

    def insertWidget(self, index: int, widget: QWidget) -> None:
        """Insert widget at specified position."""
        if index < 0:
            index = self.count()
        self.addWidget(widget)
        item = self._item_list.pop(len(self._item_list) - 1)
        self._item_list.insert(index, item)
        self.invalidate()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QPushButton

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(horz=False)
    flow_layout = QtAnimatedFlowLayout(None, use_animation=True, tight=True)
    flow_layout.setHorizontalSpacing(1)
    flow_layout.setVerticalSpacing(1)
    ha.addLayout(flow_layout)
    for text in [
        "Short",
        "Longer",
        "Different text",
        "More text",
        "Even longer button text",
        "Short",
        "Longer",
        "Different text",
        "More text",
        "Even longer button text",
    ]:
        flow_layout.addWidget(QPushButton(text))

    flow_layout = QtFlowLayout()
    flow_layout.setHorizontalSpacing(3)
    flow_layout.setVerticalSpacing(2)
    ha.addLayout(flow_layout)
    for text in [
        "Short",
        "Longer",
        "Different text",
        "More text",
        "Even longer button text",
        "Short",
        "Longer",
        "Different text",
        "More text",
        "Even longer button text",
    ]:
        flow_layout.addWidget(QPushButton(text))

    frame.show()
    sys.exit(app.exec_())



"""Loading bar."""

from qtpy.QtCore import QEasingCurve, QPropertyAnimation, QRectF, Qt
from qtpy.QtGui import QColor, QPainter
from qtpy.QtWidgets import QProgressBar


class QtLineProgressBar(QProgressBar):
    """Thin progress bar."""

    Direction = 0
    Height = 2
    Color = QColor("#00d989")
    FailedColor = QColor("#ed3814")

    # Enum attributes
    TOP = 0
    BOTTOM = 1
    Instances = {}

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(*args, parent=parent, **kwargs)
        if parent:
            parent.installEventFilter(self)
        self._height = None  # height of the widget
        self._color = None  # color of the progress bar
        self._failed_color = None  # color when progress had failed
        self._direction = None  # direction
        self._alpha = 255  # transparency
        self.is_error = False  # flag to indicate whether progress had failed
        self.setOrientation(Qt.Orientation.Horizontal)
        self.setTextVisible(False)
        self.animation = QPropertyAnimation(self, b"alpha", self, loopCount=1, duration=1000)
        self.animation.setEasingCurve(QEasingCurve.Type.SineCurve)
        self.animation.setStartValue(0)
        self.animation.setEndValue(255)
        self.Instances[self] = self

    @property
    def alpha(self):
        """Return alpha."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha
        QProgressBar.update(self)

    def paintEvent(self, _):
        """Paint event."""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform, True)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing, True)
        # background color
        painter.fillRect(self.rect(), Qt.transparent)
        # progress
        ratio = (self.value() - self.minimum()) / (self.maximum() - self.minimum())
        width = self.rect().width() * ratio
        if self.is_error:
            color = QColor(self._failed_color or QtLineProgressBar.FailedColor)
        else:
            color = QColor(self._color or QtLineProgressBar.Color)
        color.setAlpha(self._alpha)
        painter.setBrush(color)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(QRectF(0, 0, width, self.height()), 2, 2)

    def eventFilter(self, obj, event):
        """Filter events."""
        if event.type() == event.Resize:
            # resize event
            widget = QtLineProgressBar.Instances.get(obj, None)
            if widget:
                direction = widget._direction or QtLineProgressBar.Direction
                height = widget._height or QtLineProgressBar.Height
                widget.setGeometry(
                    0, 0 if direction == QtLineProgressBar.TOP else obj.height() - height, obj.width(), height
                )
        return super().eventFilter(obj, event)

    def start(self, minimum=0, maximum=100, height=None, direction=None, color=None, failed_color=None):
        """Create a loading bar.

        Parameters
        ----------
        minimum : int
            Lower step value.
        maximum : int
            Upper step value.
        height : int
            Height of the progress bar.
        direction : int
            Progress bar position
        color : str
            Success color
        failed_color : str
            Fail color.
        """
        self._height = height
        self._color = color
        self._failed_color = failed_color
        self._direction = direction
        self.setRange(minimum, maximum)
        self.setValue(minimum)
        direction = self._direction or QtLineProgressBar.Direction
        height = self._height or QtLineProgressBar.Height
        self.setGeometry(
            0,
            0 if direction == QtLineProgressBar.TOP else self.parent().height() - height,
            self.parent().width(),
            height,
        )

    def finish(self):
        """Finish."""
        self._alpha = 255
        self.is_error = False
        self.setValue(self.maximum())
        self.animation.start()

    def error(self):
        """Raise error."""
        self._alpha = 255
        self.is_error = True
        self.setValue(self.maximum())
        self.animation.start()

    def update_value(self, value):
        """Update."""
        self._alpha = 255
        self.is_error = False
        self.show()
        self.setValue(value)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        widget = QtLineProgressBar(parent=frame)
        widget.start(height=50)
        widget.update_value(50)
        ha.addWidget(widget)

        widget = QtLineProgressBar(parent=frame)
        widget.start(height=1, direction=QtLineProgressBar.TOP)
        widget.update_value(30)
        ha.addWidget(widget)

        widget = QtLineProgressBar(parent=frame)
        widget.start(height=1, direction=QtLineProgressBar.TOP)
        widget.update_value(76)
        ha.addWidget(widget)

        frame.show()
        sys.exit(app.exec_())

    _main()
